@@ -186,12 +186,14 @@ compareTest test actual = do
      then do
        modify $ \st -> st{ passed = (category test, path test) : passed st }
        -- suppress PASSED messages
-       -- liftIO $ TIO.putStrLn $ "[PASSED]   " <> path test
+       -- liftIO $ TIO.putStrLn $ "[PASSED] " <> path test
        return Passed
      else do
        modify $ \st -> st{ failed = (category test, path test) : failed st }
        liftIO $ do
-         TIO.putStrLn $ "[FAILED]   " <> T.pack (path test)
+         TIO.putStrLn $ (if path test `elem` expectedFailures
+                            then "[FAILED:EXPECTED] "
+                            else "[FAILED:UNEXPECTED] ") <> T.pack (path test)
          showDiff expected actual
        return $ Failed actual expected
 
@@ -312,7 +314,7 @@ main = do
   putStrLn ""
   let categories = sort $ Set.toList
                         $ foldr (Set.insert . category) mempty testCases
-  putStrLn $ printf "%-15s %6s %6s %6s %6s"
+  putStrLn $ printf "%-30s %6s %6s %6s %6s"
                ("CATEGORY" :: String)
                ("PASS" :: String)
                ("FAIL" :: String)
@@ -324,17 +326,17 @@ main = do
         let e = length . filter ((== cat) . fst) . errored $ counts
         let s = length . filter ((== cat) . fst) . skipped $ counts
         let percent = (fromIntegral p / fromIntegral (p + f + e) :: Double)
-        putStrLn $ printf "%-15s %6d %6d %6d %6d    |%-20s|"
+        putStrLn $ printf "%-30s %6d %6d %6d %6d    |%-20s|"
                      (T.unpack cat) p f e s
                      (replicate (floor (percent * 20.0)) '+')
   mapM_ resultsFor categories
-  putStrLn $ printf "%-15s %6s %6s %6s %6s"
+  putStrLn $ printf "%-30s %6s %6s %6s %6s"
                ("-------------" :: String)
                ("-----" :: String)
                ("-----" :: String)
                ("-----" :: String)
                ("-----" :: String)
-  putStrLn $ printf "%-15s %6d %6d %6d %6d"
+  putStrLn $ printf "%-30s %6d %6d %6d %6d"
                ("(all)" :: String)
                (length (passed counts))
                (length (failed counts))
@@ -453,5 +455,7 @@ expectedFailures = [
   "test/csl/bugreports_AutomaticallyDeleteItemsFails.txt",
   "test/csl/affix_WithCommas.txt",
   "test/csl/affix_CommaAfterQuote.txt",
-  "test/overrides/flipflop_NumericField.txt"
+  "test/overrides/flipflop_NumericField.txt",
+  "test/csl/testers_SecondAutoGeneratedZoteroPluginTest.txt",
+  "test/csl/testers_FirstAutoGeneratedZoteroPluginTest.txt"
   ]
